from collections import OrderedDict
from copy import deepcopy
import yaml
from robosuite.utils.mjcf_utils import xml_path_completion
import xml.etree.ElementTree as ET
import robocasa
import os
import re

# second keyword corresponds to positive end of axis
AXES_KEYWORDS = {0: ["left", "right"], 1: ["front", "back"], 2: ["bottom", "top"]}

# arguments not to be passed into fixture classes when initializing
IGNORE_ARGS = [
    "name",
    "align_to",
    "side",
    "alignment",
    "type",
    "center",
    "offset",
    "group_origin",
    "group_pos",
    "group_z_rot",
    "stack_height",
    "stack_fixtures",
]

# arguments used to point to other fixtures
ATTACH_ARGS = ["interior_obj", "stack_on", "attach_to"]


def initialize_fixture(config, cur_fixtures, rng=None):
    """
    initializes a fixture object based on the given configuration
    ignores positional arguments as it is changed later

    Args:
        config (dict): dictionary containing the fixture configuration.
                       Serves as the arguments to initialize the fixture

        cur_fixtures (dict): dictionary containing the current fixtures
    """

    config = deepcopy(config)
    name, class_type = config["name"], config["type"]
    if "enable" in config:
        # remove the "enable" key from the config when initializing fixture
        assert config["enable"] is True
        del config["enable"]

    # set size if stack_height is specified:
    if config.get("stack_height", None) is not None:
        stack_height = config["stack_height"]
        stack_fixtures = config["stack_fixtures"]
        curr_height = 0
        for fxtr in stack_fixtures:
            curr_height += cur_fixtures[fxtr].size[2]
        config["size"][2] = stack_height - curr_height

    # these fields should not be passed in when initializing the fixture
    for field in IGNORE_ARGS:
        if field in config:
            del config[field]

    if "pos" not in config:
        # need position to initialize fixture, adjusted later fo relative positioning
        config["pos"] = [0.0, 0.0, 0.0]

    # update fixture pointers
    for k in ATTACH_ARGS:
        if k in config:
            config[k] = cur_fixtures[config[k]]

    config["rng"] = rng
    fixture = class_type(name=name, **config)
    # print(class_type, name, type(fixture))
    return fixture


def load_style_config(style, fixture_config):
    """
    Loads the style information for a given fixture. Style information can consist of
    which xml to use if there are multiple instances of a fixture, which texture to apply,
    which subcomponents to apply (panels/handles for a cab), etc.

    Args:
        style (dict): dictionary containing the style information for each fixture type
        fixture_config (dict): dictionary containing the fixture configuration
    """
    # accounts for the different types of cabinets
    fixture_type = fixture_config["type"]

    # cabinets, shelves, drawers, and boxes use the same default configurations
    if "cabinet" in fixture_type or "drawer" in fixture_type or "box" in fixture_type:
        fixture_type = "cabinet"

    # if fixture_type not in style:
    #     raise ValueError("Did not specify fixture type \"{}\" in chosen style".format(fixture_type))
    fixture_style = style.get(fixture_type, "default")

    yaml_path = xml_path_completion(
        f"fixtures/fixture_registry/{fixture_type}.yaml",
        root=robocasa.models.assets_root,
    )
    with open(yaml_path, "r") as f:
        default_configs = yaml.safe_load(f)

    # find which configuration to use
    if isinstance(fixture_style, dict) and "config_name" not in fixture_config:
        if "default_config_name" in fixture_config:
            config_ids = fixture_style.get(
                fixture_config["default_config_name"], fixture_style["default"]
            )
            del fixture_config["default_config_name"]
        else:
            config_ids = fixture_style["default"]
    elif "default_config_name" in fixture_config:
        raise ValueError('Specified "default_config_name" but no default config found')

    elif "config_name" in fixture_config:
        config_ids = fixture_config["config_name"]
        del fixture_config["config_name"]
    else:
        config_ids = fixture_style

    # search for config by name
    config = default_configs["default"]
    if not isinstance(config_ids, list):
        config_ids = [config_ids]
    for cfg_id in config_ids:
        if cfg_id in default_configs:
            # add additional arguments based on default config
            additional_config = default_configs[cfg_id]
            for k, v in additional_config.items():
                config[k] = v
        else:
            raise ValueError(
                'Did not find style that matches "{}" for '
                'fixture type "{}"'.format(cfg_id, fixture_type)
            )
    return config


def load_wall_floor_config(layout, fixture_value):
    """
    Loads the configuration for a wall_obj and floor_obj fixture from the layout configuration.
    The configuration file path is similar to that used in load_style_config.

    Args:
        layout (dict): dictionary containing the layout information for each fixture type
        fixture_value (str): specific fixture configuration value (e.g., 'wall1')

    Returns:
        dict: The loaded wall_obj and floor_obj configuration
    """
    # Extract the fixture type
    fixture_type = fixture_value["type"]

    # Load the wall_obj configuration from the corresponding YAML file
    yaml_path = xml_path_completion(
        f"fixtures/fixture_registry/{fixture_type}.yaml",
        root=robocasa.models.assets_root,
    )
    with open(yaml_path, "r") as f:
        default_configs = yaml.safe_load(f)

    # Extract the room configuration
    room_config = layout.get("room", {})
    if fixture_type == "wall_obj":
        wall_obj_value = room_config.get("wall_obj")
    elif fixture_type == "floor_obj":
        wall_obj_value = room_config.get("floor_obj")

    # If wall_obj_value is not provided, use the fixture_value directly
    if wall_obj_value is None:
        wall_obj_value = fixture_value

    # Find which configuration to use
    if wall_obj_value not in default_configs:
        raise ValueError(
            f"Configuration '{wall_obj_value}' not found for fixture type '{fixture_type}'"
        )

    # Load the specific configuration
    config = default_configs[wall_obj_value]

    # Ensure the 'type' key is present
    if "type" not in config:
        config["type"] = fixture_type

    # Ensure the 'xml' key is present
    if "xml" not in config:
        raise ValueError(f"XML path not specified for configuration '{wall_obj_value}'")

    # Ensure the 'name' key is present
    if "name" not in config:
        config["name"] = wall_obj_value

    # Ensure the 'size' key is present
    if "size" not in config:
        config["size"] = [None, None, None]

    # Ensure the 'pos' key is present
    if "pos" not in config:
        config["pos"] = [None, None, None]

    return config


def change_fixture_style(style, layout, fixture_config):
    """
    Modifies the XML file of a fixture (wall_obj or floor_obj) based on the style configuration.

    Args:
        style (dict): dictionary containing the style information for wall and floor
        layout (dict): dictionary containing the layout information for each fixture type
        fixture_config (dict): dictionary containing the specific fixture configuration
    """
    # Extract the room configuration
    room_config = layout.get("room", {})

    # Determine the fixture kind (wall_obj or floor_obj)
    fixture_kind = fixture_config["type"]
    if fixture_kind not in ["wall_obj", "floor_obj"]:
        raise ValueError(
            f"Invalid fixture kind: {fixture_kind}. Only 'wall_obj' and 'floor_obj' are supported."
        )

    # Extract the fixture value from the room configuration
    fixture_value_key = fixture_kind
    fixture_value = room_config.get(fixture_value_key)

    # If fixture_value is not provided, use the fixture_config directly
    if fixture_value is None:
        fixture_value = fixture_config.get("name")
    # Load the fixture configuration from the corresponding YAML file
    yaml_path = xml_path_completion(
        f"fixtures/fixture_registry/{fixture_kind}.yaml",
        root=robocasa.models.assets_root,
    )

    with open(yaml_path, "r") as f:
        fixture_configs = yaml.safe_load(f)

    # Find which fixture configuration to use
    if fixture_value not in fixture_configs:
        raise ValueError(
            f"Configuration '{fixture_value}' not found for {fixture_kind}"
        )

    # Load the specific fixture configuration
    fixture_config = fixture_configs[fixture_value]
    # Ensure the 'xml' key is present
    if "xml" not in fixture_config:
        raise ValueError(f"XML path not specified for configuration '{fixture_value}'")

    # Extract the XML file path from the fixture configuration
    xml_path = xml_path_completion(
        fixture_config["xml"], root=robocasa.models.assets_root
    )

    if not xml_path.endswith("model.xml"):
        xml_path = os.path.join(xml_path, "model.xml")

    # Load the XML file
    tree = ET.parse(xml_path)
    root = tree.getroot()

    # Load the style configuration for the fixture
    style_key = fixture_kind.split("_")[0]  # Extract 'wall' or 'floor'
    fixture_style = style.get(style_key, "default")
    # Load the fixture style configuration from the corresponding YAML file
    style_yaml_path = xml_path_completion(
        f"fixtures/fixture_registry/{style_key}.yaml",
        root=robocasa.models.assets_root,
    )
    with open(style_yaml_path, "r") as f:
        fixture_styles = yaml.safe_load(f)

    # Find which fixture style configuration to use
    if fixture_style not in fixture_styles:
        raise ValueError(
            f"{style_key.capitalize()} style '{fixture_style}' not found in {style_key} configuration"
        )

    # Load the specific fixture style configuration
    fixture_style_config = fixture_styles[fixture_style]
    # Modify the XML file based on the fixture style configuration
    for texture in root.findall(".//texture"):
        texture_name = texture.get("name")
        # Check if the texture name starts with the fixture kind and ends with "_material"
        if re.match(rf"^{style_key}.*_material$", texture_name):
            if "texture" in fixture_style_config:
                # Prepend "../../../" to the texture path
                texture.set("file", f"../../../{fixture_style_config['texture']}")

    for material in root.findall(".//material"):
        material_name = material.get("name")
        # Check if the material name starts with the fixture kind and ends with "_material"
        if re.match(rf"^{style_key}.*_material$", material_name):
            if "mat_attrib" in fixture_style_config:
                for attrib, value in fixture_style_config["mat_attrib"].items():
                    material.set(attrib, value)
    # Save the modified XML file (overwrite the original file)
    tree.write(xml_path)


def get_relative_position(fixture, config, prev_fxtr, prev_fxtr_config):
    """
    Calculates the position of `fixture` based on a specified side
    and alignment relative to `prev_fixture`

    This assumes that the fixtures are properly centered!
    """

    side = config["side"].lower()
    alignment = config["alignment"].lower() if "alignment" in config else "center"
    size = fixture.size
    prev_pos, prev_size = prev_fxtr.pos, prev_fxtr.size

    # for fixtures that are not perfectly centered (e.g. stoves)
    prev_pos += prev_fxtr.origin_offset

    # place fixtures next to each others
    for axis, keywords in AXES_KEYWORDS.items():
        if side not in keywords:
            continue
        pos = deepcopy(prev_pos)
        if side == keywords[1]:
            pos[axis] = prev_pos[axis] + prev_size[axis] / 2 + size[axis] / 2
        elif side == keywords[0]:
            pos[axis] = prev_pos[axis] - prev_size[axis] / 2 - size[axis] / 2

    # align such that the specified faces are flush
    # alignment - side compatibility is checked in check_syntax()
    for axis, keywords in AXES_KEYWORDS.items():
        if keywords[0] in alignment:
            pos[axis] = prev_pos[axis] - prev_size[axis] / 2 + size[axis] / 2
        elif keywords[1] in alignment:
            pos[axis] = prev_pos[axis] + prev_size[axis] / 2 - size[axis] / 2

    if "offset" in config:
        pos += config["offset"]

    # for fixtures that are not perfectly centered (e.g. stoves)
    pos -= fixture.origin_offset
    return pos
