import os
import sys
import argparse
import numpy as np
from termcolor import colored
import matplotlib.pyplot as plt
CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

import robosuite
import robocasa
from robocasa.environments import ALL_KITCHEN_ENVIRONMENTS
from robocasa.utils.env_utils import run_random_rollouts
from robocasa.utils.gr00t_policy import load_model
from robosuite.controllers import load_composite_controller_config

#################################################################################################################################################
class ClosedLoop:

    def create_env(
        self,
        env_name,
        robots,
        controller,
        camera_names,
        render_camera,
        camera_widths,
        camera_heights,
        seed=None,
        render_onscreen=True,
        use_camera_obs = False, 
        # robocasa-related configs
        obj_instance_split=None,
        generative_textures=None,
        randomize_cameras=False,
        layout_and_style_ids=None,
        layout_ids=None,
        style_ids=None,
        translucent_robot=False,
    ):
        controller_config = load_composite_controller_config(
            controller=controller,
            robot=robots if isinstance(robots, str) else robots[0]
        )

        env_kwargs = dict(
            env_name=env_name,
            robots=robots,
            controller_configs=controller_config,
            render_camera = render_camera,
            camera_names=camera_names,
            camera_widths=camera_widths,
            camera_heights=camera_heights,
            has_renderer=render_onscreen,
            has_offscreen_renderer=use_camera_obs,
            use_camera_obs=use_camera_obs,
            ignore_done=True,
            control_freq=20,

            use_object_obs=True,
            camera_depths=False,
            seed=seed,
            obj_instance_split=obj_instance_split,
            generative_textures=generative_textures,
            randomize_cameras=randomize_cameras,
            layout_and_style_ids=layout_and_style_ids,
            layout_ids=layout_ids,
            style_ids=style_ids,
            translucent_robot=translucent_robot,
        )

        env = robosuite.make(**env_kwargs)
        return env

    #################################################################################################################################################

    def visualize_obs_state(self, step_data):
        """
        Args:
        step_data: dict()：
            'state.right_arm': shape (7,)
            'state.left_arm': shape (7,)
            'state.right_hand': shape (6,)
            'state.left_hand': shape (6,)
        Returns:
        state_vector: shape (26,) including the angles of all joints
        """
        state_right_arm = np.array(step_data["state.right_arm"]).flatten()  # shape (7,)
        state_left_arm = np.array(step_data["state.left_arm"]).flatten()  # shape (7,)
        state_right_hand = np.array(step_data["state.right_hand"]).flatten()  # shape (6,)
        state_left_hand = np.array(step_data["state.left_hand"]).flatten()  # shape (6,)

        state_vector = np.concatenate([state_right_arm, state_left_arm, state_right_hand, state_left_hand])
        return state_vector

    #################################################################################################################################################

    # create a frame_contant from last 'obs'
    def _create_frame_content(
        self, 
        obs: dict, 
        description: list[str]
    ) -> dict:
        """
        Create step_data from the last observation.

        Args:
            obs (dict): The last observation.
            description (list): The description of the action.

        Returns:
            dict: The frame content.
        """
        step_data = dict()

        # Create step_data for PandaOmron from obs
        if args.robots == "PandaOmron":
            if self.control_mode == "qpos":
                # Set Task Description
                step_data["annotation.task_index"] = description

                # Set video frame state
                step_data["video.left_view"] = np.array([np.flipud(obs["robot0_agentview_left_image"])])
                step_data["video.right_view"] = np.array([np.flipud(obs["robot0_agentview_right_image"])])
                step_data["video.wrist_view"] = np.array([np.flipud(obs["robot0_eye_in_hand_image"])])

                # Set state.arm
                cos_vals = np.array(obs["robot0_joint_pos_cos"]).flatten()
                sin_vals = np.array(obs["robot0_joint_pos_sin"]).flatten()
                angles = np.arctan2(sin_vals, cos_vals)
                step_data["state.arm"] = np.array([angles[0:7]])
                
                # Set gripper
                step_data["state.gripper"] = np.array([obs["robot0_gripper_qpos"][0]]).reshape(-1, 1)

            else:
                # Set Task Description
                step_data["annotation.human.action.task_description"] = description

                # Set video frame state
                step_data["video.left_view"] = np.array([np.flipud(obs["robot0_agentview_left_image"])])
                step_data["video.right_view"] = np.array([np.flipud(obs["robot0_agentview_right_image"])])
                step_data["video.wrist_view"] = np.array([np.flipud(obs["robot0_eye_in_hand_image"])])

                # Set End Effector relative state delta_x, delta_y, delta_z and delta_rotation_quat
                step_data["state.end_effector_position_absolute"] = np.array([obs["robot0_eef_pos"]])
                step_data["state.end_effector_rotation_absolute"] = np.array([obs["robot0_eef_quat"]])

                # Set gripper state
                step_data["state.gripper_qpos"] = np.array([obs["robot0_gripper_qpos"]])

                # Set base positon and rotation state
                step_data["state.base_position"] = np.array([obs["robot0_base_pos"]])
                step_data["state.base_rotation"] = np.array([obs["robot0_base_quat"]])

        # Create step_data for GR1ArmsOnly from obs
        elif args.robots == "GR1ArmsOnly":
            # set frame containing the annotation
            step_data["annotation.human.action.task_description"] = description

            # set frame containing the robot view image(1, 256, 256, 3)
            step_data["video.ego_view"] = np.array([np.flipud(obs["robot0_robotview_image"])])

            # set frame containing the robot state.left\right_arm
            cos_vals = np.array(obs["robot0_joint_pos_cos"]).flatten()
            sin_vals = np.array(obs["robot0_joint_pos_sin"]).flatten()
            angles = np.arctan2(sin_vals, cos_vals)
            step_data["state.right_arm"] = np.array([angles[0:7]])
            step_data["state.left_arm"] = np.array([angles[7:14]])

            # set frame containing the robot state.left\right_hand
            step_data["state.left_hand"] = np.array([[
                obs["robot0_left_gripper_qpos"][9],
                obs["robot0_left_gripper_qpos"][7],
                obs["robot0_left_gripper_qpos"][5],
                obs["robot0_left_gripper_qpos"][3],
                obs["robot0_left_gripper_qpos"][1],
                obs["robot0_left_gripper_qpos"][0],
            ]])
            step_data["state.right_hand"] = np.array([[
                obs["robot0_right_gripper_qpos"][9],
                obs["robot0_right_gripper_qpos"][7],
                obs["robot0_right_gripper_qpos"][5],
                obs["robot0_right_gripper_qpos"][3],
                obs["robot0_right_gripper_qpos"][1],
                obs["robot0_right_gripper_qpos"][0],
            ]])

        return step_data

    #################################################################################################################################################

    # convert the actions from the N1 model to the format used by robosuite
    def _act2list(self, actions: dict) -> list:
        """
        Convert the actions from the N1 model to the format used by robosuite.

        Args:
            actions (dict): The actions from the N1 model.

        Returns:
            list: The converted actions.
        """
        ordered_action = np.array([])

        if args.robots == "PandaOmron":
            if self.control_mode == "qpos":
                arm = actions["action.arm"] # (16, 7)

                gripper = actions["action.gripper"].reshape(-1, 1) # (16, 1)
                for i in range(16):
                    if gripper[i][-1] == 0:
                        gripper[i][-1] = -1
                
                base = np.zeros((16, 3)) # (16, 3)
                torso = np.zeros((16, 1)) # (16, 1)
                mode = np.zeros((16, 1)) # (16, 1)

                ordered_action = np.concatenate([arm, gripper, base, torso, mode], axis=1)

            else:
                eef_pos = actions["action.end_effector_position"] # (16, 3)
                eef_rot = actions["action.end_effector_rotation"] # (16, 3)
                gripper_close = actions["action.gripper_close"].reshape(-1, 1) # (16, 1)
                base_motion = actions["action.base_motion"] # (16, 4)
                control_mode = actions["action.control_mode"].reshape(-1, 1) # (16, 1)

                ordered_action = np.concatenate([eef_pos, eef_rot, gripper_close, base_motion, control_mode], axis=1)

        elif args.robots == "GR1ArmsOnly":
            right_arm = actions["action.right_arm"]  # (16,7)
            left_arm = actions["action.left_arm"]  # (16,7)
            right_hand = actions["action.right_hand"]  # (16,6)
            left_hand = actions["action.left_hand"]  # (16,6)

            ordered_action = np.concatenate([right_arm, left_arm, right_hand, left_hand], axis=1)

        return ordered_action

    #################################################################################################################################################

    def PandaOmron_closed_loop(self, env, lang, policy):
        counter = -50
        actions = np.array([[]])
        step_data = dict()
        action = np.array([])
        predicted_action = dict()

        print(colored(f"Instruction: {lang}", "green"))

        # Judge the controller type qpos / ik
        _, file_name = os.path.split(args.controller)
        self.control_mode = file_name.split("_")[-1][0: -5]

        if self.control_mode == "qpos":
            while counter <= 0:
                actions = np.array([[
                    0.02067039, -0.9976498, 0.01282062, -2.2387078, 0.07612378, 1.5531958, 0.7329595, # arm
                    -1, # gripper
                    0,0,0, # base
                    0, # torso
                    0, # control mode
                ]])

                obs, _, _, _ = env.step(actions[0])

                env.render()

                if counter == 0:
                    step_data = self._create_frame_content(obs, description=[lang])

                counter += 1

            while True:
                predicted_action = policy.get_action(step_data)
                actions = self._act2list(predicted_action)
                for i in range(16):
                    action = actions[i]

                    obs, _, _, _ = env.step(action)
                    env.render()

                    if i == 15:
                        step_data = self._create_frame_content(obs, description=[lang])

        else:
            while counter <= 0:
                actions = np.array([[
                    0, 0, 0,    #ee x, y, z,
                    0, 0, 0,    #ee r_x, r_y, r_z
                    0,          # gripper
                    0, 0, 0,  # base_x, base_y, base_rotation
                    0, 0        # torso, mode
                ]])
            
                obs, _, _, _ = env.step(actions[0])

                env.render()

                if counter == 0:
                    step_data = self._create_frame_content(obs, description=[lang])

                counter += 1

            while True:
                predicted_action = policy.get_action(step_data)
                actions = self._act2list(predicted_action)
                for i in range(5):
                    action = actions[i]

                    obs, _, _, _ = env.step(action)
                    env.render()

                    if i == 4:
                        step_data = self._create_frame_content(obs, description=[lang])

    #################################################################################################################################################

    def GR1ArmsOnly_closed_loop(self, env, lang, policy):

        # open interactive mode plt.ion()
        joint_names = [
            "r_shoulder_pitch","r_shoulder_roll","r_shoulder_yaw","r_elbow","r_wrist_yaw","r_wrist_roll","r_wrist_pitch",
            "l_shoulder_pitch","l_shoulder_roll","l_shoulder_yaw","l_elbow","l_wrist_yaw","l_wrist_roll","l_wrist_pitch",
            "r_pinky","r_ring","r_middle","r_index","r_thumb_pitch","r_thumb_yaw",
            "l_pinky","l_ring","l_middle","l_index","l_thumb_pitch","l_thumb_yaw",
        ]
        x = np.arange(len(joint_names))
        plt.ion()
        fig, ax = plt.subplots(figsize=(12, 6))
        (line_command,) = ax.plot(x, np.zeros(26), "o-", label="Commanded Action")
        (line_state,) = ax.plot(x, np.zeros(26), "s-", label="Observed State")
        ax.set_xticks(x)
        ax.set_xticklabels(joint_names, rotation=45, fontsize=10)
        ax.set_xlabel("Joint")
        ax.set_ylabel("Angle (radians)")
        ax.set_title("Commanded Action vs Observed Joint Angles")
        ax.legend()
        plt.show()

        counter = -50
        step_data = dict()
        predicted_action = dict()
        actions = list()
        action = list()
        test_step = dict()

        print(colored(f"Instruction: {lang}", "green"))

        while counter <= 0:
            actions = np.array([[
                0.0,0.0 ,0.0,-1.57,0.0,0.0,0.0,
                0.0,0.0 ,0.0,-1.57,0.0,0.0,0.0,
                -1.5,-1.5,-1.5,-1.5,-3.0, 3.0,
                -1.5,-1.5,-1.5,-1.5,-3.0, 3.0,
            ]])

            obs, reward, done, info = env.step(actions[0])  # take action in the environment
            env.render()  # render on display

            if counter == 0:
                step_data = self._create_frame_content(obs, description=[lang])

            counter += 1

        while True:
            predicted_action = policy.get_action(step_data)
            actions = self._act2list(predicted_action)
            for i in range(5):
                """
                action[i](26, ): [
                    r_should_pitch, r_should_roll, r_should_yaw, r_elbow, r_wrist_yaw, r_wrist_roll, r_wrist_pitch,
                    l_should_pitch, l_should_roll, l_should_yaw, l_elbow, l_wrist_yaw, l_wrist_roll, l_wrist_pitch,
                    r_pinky, r_ring, r_middle, r_index, r_thumb_pitch, r_thumb_yaw,
                    l_pinky, l_ring, l_middle, l_index, l_thumb_pitch, l_thumb_yaw
                ]
                """
                action = actions[i]

                obs, reward, done, info = env.step(action)  # take action in the environment
                env.render()  # render on display

                # Draw control qpos signal and actural qpos in simulaor
                test_step = self._create_frame_content(obs, description=[lang])
                line_command.set_ydata(action)
                line_state.set_ydata(self.visualize_obs_state(test_step))
                ax.relim()
                ax.autoscale_view()
                plt.pause(0.1)

                if i == 4:
                    step_data = self._create_frame_content(obs, description=[lang])

    #################################################################################################################################################

    def H1ArmsOnly_closed_loop(self, env, lang):
        while True:
            # Take a default action in the environment
            action = np.array([0, 0, 0, 0,
                                0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0])
            obs, reward, done, info = env.step(action)
            env.render()  # render on display


    def G1ArmsOnly_closed_loop(self, env, lang):
        while True:
            # Take a default action in the environment
            action = np.array([0, 0, 0, 1.57, 0, 0, 0,
                                0, 0, 0, 1.57, 0, 0, 0,
                                0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0])
            obs, reward, done, info = env.step(action)
            env.render()  # render on display

    #################################################################################################################################################

def main():
    loop = ClosedLoop()
    # Create environment: env
    env = loop.create_env(
        env_name = args.env_name,
        robots = args.robots,
        controller = args.controller,
        camera_names = args.camera_names,
        use_camera_obs = args.use_camera_obs,
        render_camera= args.render_camera,
        render_onscreen = args.render_onscreen,
        camera_widths = args.camera_widths,
        camera_heights = args.camera_heights,
        # robocasa-related configs
        translucent_robot=args.translucent_robot,
    )

    # Get environment task description: "lang"
    ep_meta = env.get_ep_meta()
    lang = ep_meta.get("lang", None)

    # Load Policy: policy
    dataset, policy = load_model(args.model_config, args.model_tag, args.model_path, args.dataset_path)

    # reset env
    env.reset()

    # ClosedLoop with different robots
    if args.robots == "PandaOmron":
        loop.PandaOmron_closed_loop(env, lang, policy)
    elif args.robots == "GR1ArmsOnly":
        loop.GR1ArmsOnly_closed_loop(env, lang, policy)
    elif args.robots == "H1ArmsOnly":
        loop.H1ArmsOnly_closed_loop(env, lang)
    elif args.robots == "G1ArmsOnly":
        loop.G1ArmsOnly_closed_loop(env, lang)

#################################################################################################################################################

if __name__ == "__main__":

    # Arguments
    parser = argparse.ArgumentParser()

    # All default is for PandaOmron CloseDrawer Task
    parser.add_argument("-e", "--env_name", type=str, 
                        default="PnPCounterToStove", 
                        help="Which task to use")
    parser.add_argument("-b", "--robots", type=str, 
                        default="PandaOmron", 
                        help="Which robot(s) to use in the env")
    parser.add_argument("--controller", type=str, 
                        default="robosuite/robosuite/controllers/config/robots/default_pandaomron_qpos.json", 
                        help="Choice of controller. Can be, eg. 'NONE' or 'WHOLE_BODY_IK', etc. Or path to controller json file")
    parser.add_argument("-r", "--render_onscreen", type=bool, 
                        default=True, 
                        help="Whether to render on screen")
    parser.add_argument("--render_camera", type=str, 
                        default="robot0_agentview_left", 
                        help="Which camera to render the environment")
    parser.add_argument("-o", "--use_camera_obs", type=bool, 
                        default=True, 
                        help="Whether to use camera for obs(returned by env.step)")
    parser.add_argument("-c", "--camera_names", type=str, 
                        default=["robot0_agentview_left", "robot0_agentview_right", "robot0_eye_in_hand"], 
                        help="Which camera to use for obs(returned by env.step)")    
    parser.add_argument("--camera_widths", type=int, 
                        default=[224, 224, 224], 
                        help="Cameras' widths")
    parser.add_argument("--camera_heights", type=int, 
                        default=[224, 224, 224], 
                        help="Cameras' heights")
    parser.add_argument("-t", "--translucent_robot", type=bool, 
                        default=False, 
                        help="Whether set the robot translucent")
    parser.add_argument("--model_tag", type=str, 
                        default="new_embodiment", 
                        help="'gr1' or 'new_embodiment'")
    parser.add_argument("--model_config", type=str, 
                        default="single_panda_gripper2", 
                        help=("[gr1_arms_only, gr1_arms_waist, gr1_full_upper_body, "
                              "bimanual_panda_gripper,bimanual_panda_hand, single_panda_gripper, so100]"))
    parser.add_argument("--model_path", type=str, 
                        default="/home/szliutong/PandaOmron_N1_pnp/checkpoint-100000", 
                        help="Model path of GR00T N1")
    parser.add_argument("--dataset_path", type=str, 
                        default="/home/szliutong/dataset/robocasa/PnPCounterToStoveMG", 
                        help="Dataset path of GR00T N1")
    args = parser.parse_args()

    main()