"""
A script to collect a batch of human demonstrations that can be used
to generate a learning curriculum (see `demo_learning_curriculum.py`).

The demonstrations can be played back using the `playback_demonstrations_from_pkl.py`
script.
"""

import argparse
from copy import deepcopy
import datetime
import json
import os
import time
from glob import glob
import sys
import tarfile
import subprocess
import h5py
import imageio
import mujoco
import numpy as np
import platform
import robosuite
import traceback
import re

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

# from robosuite import load_controller_config
from robosuite.controllers import load_composite_controller_config
from robosuite.wrappers import DataCollectionWrapper, VisualizationWrapper
from termcolor import colored

os.environ["ROBOCASA_ASSETS_ROOT"] = os.path.join(
    os.path.abspath(os.path.dirname(os.path.dirname(__file__))), "models", "assets"
)
import robocasa
import robocasa.macros as macros
from robocasa.models.fixtures import FixtureType
from robocasa.utils.robomimic.robomimic_dataset_utils import convert_to_robomimic_format
from multiprocessing import shared_memory
from robocasa_upload.UploadGenerator import UploadGenerator

DELTA_INPUT_DEVICES = ["spacemouse"]

DEVICE_MAPPING = {
    "keyboard": ("robosuite.devices", "Keyboard"),
    "spacemouse": ("robosuite.devices", "SpaceMouse"),
    "vr-controller": ("robosuite.devices.vr", "VRController"),
    "vr-hand": ("robosuite.devices.vr", "VRHand"),
}

def strip_color(text):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)

def parse_dict(arg):
    try:
        # 将字符串转换为字典
        # 支持格式: "env_name:CheesyBread,layout_ids:6"
        if not arg:
            return {}
        pairs = arg.split(',')
        return dict(pair.split(':') for pair in pairs)
    except:
        raise argparse.ArgumentTypeError('必须是 key:value 格式，多个键值对用逗号分隔')

def collect_human_trajectory(
    env,
    device,
    arm,
    env_configuration,
    mirror_actions,
    render=True,
    max_fr=None,
    print_info=True,
):
    """
    Use the device (keyboard or SpaceNav 3D mouse) to collect a demonstration.
    The rollout trajectory is saved to files in npz format.
    Modify the DataCollectionWrapper wrapper to add new fields or change data formats.

    Args:
        env (MujocoEnv): environment to control
        device (Device): to receive controls from the device
        arms (str): which arm to control (eg bimanual) 'right' or 'left'
        env_configuration (str): specified environment configuration
    """

    env.reset()

    ep_meta = env.get_ep_meta()
    # print(json.dumps(ep_meta, indent=4))
    lang = ep_meta.get("lang", None)
    if print_info and lang is not None:
        print(colored(f"Instruction: {lang}", "green"))

    # degugging: code block here to quickly test and close env
    # env.close()
    # return None, True

    if render:
        # ID = 2 always corresponds to agentview
        env.render()

    task_completion_hold_count = (
        -1
    )  # counter to collect 10 timesteps after reaching goal
    device.start_control()

    nonzero_ac_seen = False

    # Keep track of prev gripper actions when using since they are position-based and must be maintained when arms switched
    all_prev_gripper_actions = [
        {
            f"{robot_arm}_gripper": np.repeat([0], robot.gripper[robot_arm].dof)
            for robot_arm in robot.arms
            if robot.gripper[robot_arm].dof > 0
        }
        for robot in env.robots
    ]

    # H1ArmsOnly: 'right'(0, 6) 'left'(6, 12) 'right_gripper'(12, 18) 'left_gripper'(18, 24)
    # GR1FloatingBody: 'head'(0, 6) 'right'(6, 12) 'left'(12, 18) 'base'(18, 21) 'right_gripper'(21, 27) 'left_gripper'(27, 33)
    zero_action = np.zeros(env.action_dim)
    for _ in range(1):
        # do a dummy step thru base env to initalize things, but don't record the step
        if isinstance(env, DataCollectionWrapper):
            env.env.step(zero_action)
        else:
            env.step(zero_action)

    discard_traj = False
    reset_status = False

    # Loop until we get a reset from the input or the task completes
    while True:
        start = time.time()

        # Set active robot
        active_robot = env.robots[device.active_robot]
        active_arm = device.active_arm

        # Get the newest action
        input_ac_dict = device.input2action(mirror_actions=mirror_actions)

        # If action is none, then this a reset so we should break
        if input_ac_dict is None:
            discard_traj = True
            reset_status = True
            break

        action_dict = deepcopy(input_ac_dict)

        if "head" in input_ac_dict:
            action_dict["head"] = input_ac_dict["head"]

        # set arm actions
        for arm in active_robot.arms:
            controller_input_type = active_robot.part_controllers[arm].input_type
            if controller_input_type == "delta":
                action_dict[arm] = input_ac_dict[f"{arm}_delta"]
            elif controller_input_type == "absolute":
                action_dict[arm] = input_ac_dict[f"{arm}_abs"]
            else:
                raise ValueError
        # print(action_dict)
        if device.is_empty_input(action_dict):
            if not nonzero_ac_seen:
                if render:
                    env.render()
                continue
        else:
            nonzero_ac_seen = True

        # Maintain gripper state for each robot but only update the active robot with action
        env_action = [
            robot.create_action_vector(all_prev_gripper_actions[i])
            for i, robot in enumerate(env.robots)
        ]
        env_action[device.active_robot] = active_robot.create_action_vector(action_dict)
        env_action = np.concatenate(env_action)

        # Run environment step
        obs, _, _, _ = env.step(env_action)
        if render:
            env.render()

        # Also break if we complete the task
        if task_completion_hold_count == 0:
            break

        # state machine to check for having a success for 15 consecutive timesteps
        if env._check_success():
            if task_completion_hold_count > 0:
                task_completion_hold_count -= 1  # latched state, decrement count
            else:
                task_completion_hold_count = 15  # reset count on first success timestep
        else:
            task_completion_hold_count = -1  # null the counter if there's no success

        # limit frame rate if necessary
        if max_fr is not None:
            elapsed = time.time() - start
            diff = 1 / max_fr - elapsed
            if diff > 0:
                time.sleep(diff)

        # with open("/home/soroushn/tmp/model.xml", "w") as f:
        #     f.write(env.model.get_xml())
        # exit()

    if nonzero_ac_seen and hasattr(env, "ep_directory"):
        ep_directory = env.ep_directory
    else:
        ep_directory = None

    # cleanup for end of data collection episodes
    # if close here, hard_reset must be True!!(otherwise self.sim will be None, reset whill occur bug)
    env._flush()

    return ep_directory, discard_traj, reset_status


def gather_demonstrations_as_hdf5(
    directory,
    out_dir,
    env_info,
    successful_episodes=None,
    verbose=False,
    out_name="demo.hdf5",
):
    """
    Gathers the demonstrations saved in @directory into a
    single hdf5 file.
    The strucure of the hdf5 file is as follows.
    data (group)
        date (attribute) - date of collection
        time (attribute) - time of collection
        repository_version (attribute) - repository version used during collection
        env (attribute) - environment name on which demos were collected
        demo1 (group) - every demonstration has a group
            model_file (attribute) - model xml string for demonstration
            states (dataset) - flattened mujoco states
            actions (dataset) - actions applied during demonstration
        demo2 (group)
        ...
    Args:
        directory (str): Path to the directory containing raw demonstrations.
        out_dir (str): Path to where to store the hdf5 file.
        env_info (str): JSON-encoded string containing environment information,
            including controller and robot info
    """

    hdf5_path = os.path.join(out_dir, out_name)
    if verbose:
        print("Saving hdf5 to", hdf5_path)
    f = h5py.File(hdf5_path, "w")

    # store some metadata in the attributes of one group
    grp = f.create_group("data")

    num_eps = 0
    env_name = None  # will get populated at some point

    for ep_directory in os.listdir(directory):
        # print("Processing {} ...".format(ep_directory))
        if (successful_episodes is not None) and (
            ep_directory not in successful_episodes
        ):
            # print("\tExcluding this episode!")
            continue

        state_paths = os.path.join(directory, ep_directory, "state_*.npz")
        states = []
        actions = []
        actions_abs = []
        # success = False

        for state_file in sorted(glob(state_paths)):
            dic = np.load(state_file, allow_pickle=True)
            env_name = str(dic["env"])

            states.extend(dic["states"])
            for ai in dic["action_infos"]:
                actions.append(ai["actions"])
                if "actions_abs" in ai:
                    actions_abs.append(ai["actions_abs"])
            # success = success or dic["successful"]

        if len(states) == 0:
            continue

        # # Add only the successful demonstration to dataset
        # if success:

        # print("Demonstration is successful and has been saved")
        # Delete the last state. This is because when the DataCollector wrapper
        # recorded the states and actions, the states were recorded AFTER playing that action,
        # so we end up with an extra state at the end.
        del states[-1]
        assert len(states) == len(actions)

        num_eps += 1
        ep_data_grp = grp.create_group("demo_{}".format(num_eps))

        # store model xml as an attribute
        xml_path = os.path.join(directory, ep_directory, "model.xml")
        with open(xml_path, "r") as f:
            xml_str = f.read()
        ep_data_grp.attrs["model_file"] = xml_str

        # store ep meta as an attribute
        ep_meta_path = os.path.join(directory, ep_directory, "ep_meta.json")
        if os.path.exists(ep_meta_path):
            with open(ep_meta_path, "r") as f:
                ep_meta = f.read()
            ep_data_grp.attrs["ep_meta"] = ep_meta

        # write datasets for states and actions
        ep_data_grp.create_dataset("states", data=np.array(states))
        ep_data_grp.create_dataset("actions", data=np.array(actions))
        if len(actions_abs) > 0:
            print(np.array(actions_abs).shape)
            ep_data_grp.create_dataset("actions_abs", data=np.array(actions_abs))

        # else:
        #     pass
        #     # print("Demonstration is unsuccessful and has NOT been saved")

    if verbose:
        print("{} successful demos so far".format(num_eps))

    if num_eps == 0:
        f.close()
        return

    # write dataset attributes (metadata)
    now = datetime.datetime.now()
    grp.attrs["date"] = "{}-{}-{}".format(now.month, now.day, now.year)
    grp.attrs["time"] = "{}:{}:{}".format(now.hour, now.minute, now.second)
    grp.attrs["robocasa_version"] = robocasa.__version__
    grp.attrs["robosuite_version"] = robosuite.__version__
    grp.attrs["mujoco_version"] = mujoco.__version__
    grp.attrs["env"] = env_name
    grp.attrs["env_info"] = env_info

    f.close()

    return hdf5_path

def initialize_device(env, config):
    device_type = config["device"]
    pos_sensitivity = config["pos_sensitivity"]
    rot_sensitivity = config["rot_sensitivity"]
    is_continuous_pressing = config["is_continuous_pressing"]
    if device_type not in DEVICE_MAPPING:
        raise ValueError(f"Invalid device type: {device_type}")
        
    module_path, class_name = DEVICE_MAPPING[device_type]
    device_module = __import__(module_path, fromlist=[class_name])
    device_class = getattr(device_module, class_name.split(".")[-1])

    absolute_device_params = {
        "env": env,
    }

    delta_device_params = {
        **absolute_device_params,
        "pos_sensitivity": pos_sensitivity,
        "rot_sensitivity": rot_sensitivity,
    }
    
    if device_type == "keyboard":
        device_params = delta_device_params

    elif device_type == "spacemouse":
        device_params = {
            **delta_device_params,
            "vendor_id": macros.SPACEMOUSE_VENDOR_ID,
            "product_id": macros.SPACEMOUSE_PRODUCT_ID,
            "is_continuous_pressing": is_continuous_pressing}

    elif device_type.startswith("vr"):
        env.unwrapped.reset()
        image_size = (720, 1280)
        shm = shared_memory.SharedMemory(
            create=True,
            size=image_size[0] * image_size[1] * 3 * np.uint8().itemsize,
        )
        device_params = {
            **absolute_device_params,
            "img_shape": image_size,
            "shm_name": shm.name,
            "robot_mode": "arm"}

    return device_class(**device_params), shm if device_type.startswith("vr") else None

def create_env_wrapper(env_config, experiment_config):
    env_config["controller_configs"] = load_composite_controller_config(
        controller=experiment_config["controller"],
        robot=env_config["robots"] if isinstance(env_config["robots"], str) else env_config["robots"][0],
    )
    if experiment_config["device"] in DELTA_INPUT_DEVICES:
        env_config["controller_configs"]["body_parts"]["right"]["input_type"] = "delta"
        print(colored(f"Using {experiment_config['device']}. Set delta input type for arms", "red"))
    if env_config["controller_configs"]["type"] == "WHOLE_BODY_MINK_IK":
        # mink-speicific import. requires installing mink
        from robosuite.examples.third_party_controller.mink_controller import (
            WholeBodyMinkIK,
        )
    env = robosuite.make(
         **env_config,
         has_renderer=True,
         has_offscreen_renderer=False,
         render_camera=experiment_config["camera"],
         ignore_done=True,
         use_camera_obs=False,
         control_freq=20,
         renderer=experiment_config["renderer"],
    )
    origin_layout_and_style_ids = env.layout_and_style_ids
    # Wrap this with visualization wrapper
    env = VisualizationWrapper(env)
    # Grab reference to controller config and convert it to json-encoded string
    env_info = json.dumps(env_config)
    t_now = time.time()
    time_str = datetime.datetime.fromtimestamp(t_now).strftime("%Y-%m-%d-%H-%M-%S")
    time_str = f"{time_str}_{env_config['env_name']}"
    # make a new timestamped directory
    demo_dir = os.path.join(experiment_config["directory"], time_str)
    os.makedirs(demo_dir, exist_ok=True)
    # wrap the environment with data collection wrapper
    all_eps_directory = os.path.join(demo_dir, "episodes")
    env = DataCollectionWrapper(env, all_eps_directory)
    return env, env_info, origin_layout_and_style_ids, demo_dir, all_eps_directory

def task_valid_judge(env, upload_gen, demo_dir, env_info, metadata, error_info="", error_desc=""):
    if metadata["error_type"] == "code":
        upload_gen.upload_error(error_info, env_config["env_name"], metadata, desc=error_desc)
        upload_gen.update_task(mode="badtask")
        return
    user_input = input(colored(f"Is this task bad? (y/n)", "red"))
    while True:
        if user_input.lower() == 'y':
            description = input(colored(f"Why:", "red"))
            while True:
                if(len(description)<=1):
                    print(colored("Invalid input, please input a description:", "red"))
                    description = input()
                else:
                    break
            if(
                env is not None
                and ep_layout_id is not None
                and env.ep_directory is not None
            ):
                upload_gen.upload(demo_dir,
                                  os.path.split(env.ep_directory)[-1],
                                  env_config["env_name"],
                                  env_info,
                                  metadata,
                                  desc=description
                )
            upload_gen.update_task(mode="badtask")
            break
        elif user_input.lower() == 'n':
            upload_gen.update_task(mode="reset")
            break
        else:
            print(colored("Invalid input, please input y/n:", "red"))
            user_input = input()

if __name__ == "__main__":
    # Arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--directory",
        type=str,
        default=os.path.join(robocasa.models.assets_root, "demonstrations_private"),
    )
    parser.add_argument("--environment", type=str, default="Kitchen")
    parser.add_argument(
        "--robots",
        nargs="+",
        type=str,
        default="PandaOmron",
        help="Which robot(s) to use in the env",
    )
    parser.add_argument(
        "--config",
        type=str,
        default="single-arm-opposed",
        help="Specified environment configuration if necessary",
    )
    parser.add_argument(
        "--arm",
        type=str,
        default="right",
        help="Which arm to control (eg bimanual) 'right' or 'left'",
    )
    parser.add_argument(
        "--obj_groups",
        type=str,
        nargs="+",
        default=None,
        help="In kitchen environments, either the name of a group to sample object from or path to an .xml file",
    )

    parser.add_argument(
        "--camera",
        type=str,
        default=None,
        help="Which camera to use for collecting demos",
    )
    parser.add_argument(
        "--controller",
        type=str,
        default=None,
        help="Choice of controller. Can be, eg. 'NONE' or 'WHOLE_BODY_IK', etc. Or path to controller json file",
    )
    parser.add_argument(
        "--device",
        type=str,
        default="spacemouse",
        choices=["keyboard", "spacemouse", "vr-controller", "vr-hand"],
    )
    parser.add_argument(
        "--pos-sensitivity",
        type=float,
        default=4.0,
        help="How much to scale position user inputs",
    )
    parser.add_argument(
        "--rot-sensitivity",
        type=float,
        default=4.0,
        help="How much to scale rotation user inputs",
    )
    parser.add_argument(
        "--continuous-pressing",
        type=bool,
        default=False,
        help="Whether to use continuous pressing for spacemouse",
    )
    parser.add_argument(
        "--project",
        type=str,
        default="",
        help="Which project to collect",
    )

    parser.add_argument("--debug", action="store_true")
    parser.add_argument(
        "--renderer", type=str, default="mjviewer", choices=["mjviewer", "mujoco"]
    )
    parser.add_argument(
        "--max_fr", default=30, type=int, help="If specified, limit the frame rate"
    )

    parser.add_argument("--layout", type=int, nargs="+", default=-1)
    parser.add_argument(
        "--style",
        type=int,
        nargs="+",
        default=[101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113],
    )
    parser.add_argument("--auto", action="store_true")
    parser.add_argument("--batch_name", type=str, default="")
    parser.add_argument("--filters", type=parse_dict, default={})
    parser.add_argument("--generative_textures", action="store_true")
    parser.add_argument("--obj_fix", action="store_true")
    args = parser.parse_args()

    # Create argument configuration
    env_config = {
        "env_name": args.environment,
        "robots": args.robots,
    }

    if args.generative_textures is True:
        env_config["generative_textures"] = "100p"

    # Check if we're using a multi-armed environment and use env_configuration argument if so
    if "TwoArm" in args.environment:
        env_config["env_configuration"] = args.config

    # Mirror actions if using a kitchen environment
    if args.environment in ["Lift"]:  # add other non-kitchen tasks here
        if args.obj_groups is not None:
            print(
                "Specifying 'obj_groups' in non-kitchen environment does not have an effect."
            )
        mirror_actions = False
        if args.camera is None:
            args.camera = "agentview"
        # special logic: "free" camera corresponds to Null camera
        elif args.camera == "free":
            args.camera = None
    else:
        mirror_actions = True
        if args.layout is not None:
            env_config["layout_ids"] = args.layout
        if args.style is not None:
            env_config["style_ids"] = args.style
        ### update config for kitchen envs ###
        if args.obj_groups is not None:
            env_config.update({"obj_groups": args.obj_groups})
        if args.camera is None:
            args.camera = "robot0_frontview"
        # special logic: "free" camera corresponds to Null camera
        elif args.camera == "free":
            args.camera = None

        env_config["translucent_robot"] = True

        # by default use obj instance split A
        env_config["obj_instance_split"] = "A"
        # config["obj_instance_split"] = None
        # config["obj_registries"] = ("aigen",)

        if args.obj_fix:
            env_config["seed"] = 729
            env_config["hard_reset"] = False

    experiment_config = {
        "controller": args.controller,
        "device": args.device,
        "camera": args.camera,
        "renderer": args.renderer,
        "directory": args.directory,
        "pos_sensitivity": args.pos_sensitivity,
        "rot_sensitivity": args.rot_sensitivity,
        "is_continuous_pressing": args.continuous_pressing,
        "project_name": args.project,
    }

    if platform.system() == "Windows":
        os.environ["MUJOCO_GL"] = "wgl"
    elif platform.system() == "Darwin":
        os.environ["MUJOCO_GL"] = "cgl"
    elif platform.system() == "Linux":
        os.environ["MUJOCO_GL"] = "egl"

    # init upload generator
    os.environ['PYTHONOPTIMIZE'] = '1'  # 启用Python优化
    upload_gen = UploadGenerator(args.directory, args.project)

    # init metadata
    metadata = None

    # Create environment
    if not args.auto:
        metadata = {
            'robot_name': env_config["robots"] if isinstance(env_config["robots"], str) else env_config["robots"][0],
            'device_name': experiment_config["device"],
            'system': platform.system(),
            'project_name': experiment_config["project_name"]
        }
        (
            env,
            env_info,
            origin_layout_and_style_ids,
            demo_dir,
            all_eps_directory,
        ) = create_env_wrapper(env_config, experiment_config)
        device, shm = initialize_device(env, experiment_config)
    else:
        env = None
        env_info = None
        origin_layout_and_style_ids = None
        demo_dir = None
        all_eps_directory = None
        device = None

    successful_episodes = []

    reset_status = False
    # collect demonstrations
    try:
        while True:
            print()
            # 拉取任务
            if args.auto and not reset_status:
                task_info = upload_gen.pull_task(batch_name=args.batch_name, filters=args.filters)
                if env is not None:
                    env.close()
                    env = None
                    metadata = None
                for key, value in task_info.items(): 
                    if key in list(env_config.keys()):
                        env_config[key] = value
                    if key in list(experiment_config.keys()):
                        experiment_config[key] = value
                metadata = {
                    'robot_name': env_config["robots"] if isinstance(env_config["robots"], str) else env_config["robots"][0],
                    'device_name': experiment_config["device"],
                    'system': platform.system(),
                    'project_name': experiment_config["project_name"],
                }
                (
                    env,
                    env_info,
                    origin_layout_and_style_ids,
                    demo_dir,
                    all_eps_directory,
                ) = create_env_wrapper(env_config, experiment_config)
                if device is not None:
                    device.close()
                device, shm = initialize_device(env, experiment_config)

            ep_directory, discard_traj, reset_status = collect_human_trajectory(
                env,
                device,
                args.arm,
                args.config,
                mirror_actions,
                render=(args.renderer != "mjviewer"),
                max_fr=args.max_fr,
            )

            ep_layout_id = None
            if os.path.exists(os.path.join(ep_directory, "ep_meta.json")):
                with open(os.path.join(ep_directory, "ep_meta.json"), "r") as f:
                    json_content = json.load(f)
                    ep_layout_id = json_content["layout_id"]
                    ep_style_id = json_content["style_id"]
            elif args.auto:
                ep_layout_id = task_info["layout_ids"]
            
            metadata.update({'layout_id': ep_layout_id, 'style_id': ep_style_id, "is_success": not discard_traj})

            # if reset, keep layout
            if reset_status and ep_layout_id is not None:
                # DataCollectionWrapper -> VisualizationWrapper -> MujocoEnv
                env.env.env.layout_and_style_ids = [
                    pair
                    for pair in origin_layout_and_style_ids
                    if pair[0] == ep_layout_id
                ]
            else:
                env.env.env.layout_and_style_ids = origin_layout_and_style_ids

            if ep_directory is not None:
                # save whether the episode is successful or not
                with open(os.path.join(ep_directory, "ep_stats.json"), "w") as file:
                    json.dump({"success": not discard_traj}, file)

                with open(os.path.join(ep_directory, "env_info.json"), "w") as file:
                    json.dump(env_info, file)

                if not discard_traj:
                    successful_episodes.append(os.path.split(ep_directory)[-1])

                gather_demonstrations_as_hdf5(
                    all_eps_directory,
                    ep_directory,
                    env_info,
                    successful_episodes=[os.path.split(ep_directory)[-1]],
                    out_name="ep_demo.hdf5",
                )

            print("Episode success:", not discard_traj)

            if not args.debug:
                hdf5_path = gather_demonstrations_as_hdf5(
                    all_eps_directory,
                    demo_dir,
                    env_info,
                    successful_episodes=successful_episodes,
                    verbose=True,
                )
                if hdf5_path is not None:
                    convert_to_robomimic_format(hdf5_path)

            if upload_gen is not None and ep_layout_id is not None:
                user_input = input(colored("Upload this episode? (y/n): ", "cyan"))
                while True:
                    if user_input.lower() == 'n':
                        reset_status = True
                        break
                    elif user_input.lower() == 'y':
                        upload_gen.upload(demo_dir, os.path.split(ep_directory)[-1], env_config["env_name"], env_info, metadata)
                        reset_status = False # 重置reset状态
                        break
                    else:
                        print(colored("Invalid input, please input y/n:", "red"))
                        user_input = input()

    except KeyboardInterrupt:
        print("\nKeyboard interrupt detected. Performing cleanup...")
        if env is not None and env.ep_directory is not None:
            gather_demonstrations_as_hdf5(
                all_eps_directory,
                env.ep_directory,
                env_info,
                successful_episodes=[os.path.split(env.ep_directory)[-1]],
                out_name="ep_demo.hdf5",
            )
        if os.path.exists(os.path.join(env.ep_directory, "ep_meta.json")):
            with open(os.path.join(env.ep_directory, "ep_meta.json"), "r") as f:
                json_content = json.load(f)
                ep_layout_id = json_content["layout_id"]
                ep_style_id = json_content["style_id"]
        elif args.auto:
            ep_layout_id = task_info["layout_ids"]
        metadata.update({"error_type": "user", "layout_id": ep_layout_id, "is_success": False})
        task_valid_judge(env, upload_gen, demo_dir, env_info, metadata)

    except Exception as e:
        print(colored(traceback.format_exc(), "red"))
        if strip_color(str(e)) != "No task left":
            metadata.update({"error_type": "code", "layout_id": task_info["layout_ids"], "is_success": False})
            task_valid_judge(
                env,
                upload_gen,
                demo_dir,
                env_info,
                metadata,
                error_info=traceback.format_exc(),
                error_desc=f"Code Error: {e}"
            )

    finally:
        if args.device.startswith("vr"):
            device.close()
            shm.unlink()
            shm.close()
        if env is not None:
            env.close()
    sys.exit(0)
