from lxml import etree
import os
os.environ["ROBOCASA_ASSETS_ROOT"] = os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))), "models", "assets")
import glob
import numpy as np
from robocasa.models.objects.objects import MujocoXMLObjectRobocasa
from robosuite.utils.mjcf_utils import string_to_array, array_to_string
from tqdm import tqdm
import time
import xml.etree.ElementTree as ET
import robosuite
import robosuite.utils.transform_utils as T
import argparse


BASE_PATH = "/robocasa/Robocasa_LW_Objects_20250519/Objects/sandwich_breads"
GEOM_TEMPLATE = '<geom class="region" name="reg_bbox" type="box" pos="{}" size="{}"/>'
REGION_GEOM_CLS = '''
<default>
    <default class="region">
        <geom group="1" conaffinity="0" contype="0" rgba="0 1 0 0"/>
    </default>
</default>
'''

class MJCFSiteObject(MujocoXMLObjectRobocasa):
    """
    Blender object with support for changing the scaling
    """

    def __init__(
        self,
        name,
        mjcf_path,
        scale=1.0,
        solimp=(0.998, 0.998, 0.001),
        solref=(0.001, 1),
        density=100,
        friction=(0.95, 0.3, 0.1),
        margin=None,
        rgba=None,
        priority=None,
    ):
        # get scale in x, y, z
        if isinstance(scale, float):
            scale = [scale, scale, scale]
        elif isinstance(scale, tuple) or isinstance(scale, list):
            assert len(scale) == 3
            scale = tuple(scale)
        else:
            raise Exception("got invalid scale: {}".format(scale))
        scale = np.array(scale)

        self.solimp = solimp
        self.solref = solref
        self.density = density
        self.friction = friction
        self.margin = margin

        self.priority = priority

        self.rgba = rgba

        # read default xml
        xml_path = mjcf_path
        folder = os.path.dirname(xml_path)
        tree = ET.parse(xml_path)
        root = tree.getroot()

        # write modified xml (and make sure to postprocess any paths just in case)
        xml_str = ET.tostring(root, encoding="utf8").decode("utf8")
        xml_str = self.postprocess_model_xml(xml_str)
        time_str = str(time.time()).replace(".", "_")
        new_xml_path = os.path.join(folder, "{}_{}.xml".format(time_str, os.getpid()))
        f = open(new_xml_path, "w")
        f.write(xml_str)
        f.close()

        # initialize object with new xml we wrote
        super().__init__(
            fname=new_xml_path,
            name=name,
            joints=[dict(type="free", damping="0.0005")],
            obj_type="all",
            duplicate_collision_geoms=False,
            scale=scale,
        )

        # clean up xml - we don't need it anymore
        if os.path.exists(new_xml_path):
            os.remove(new_xml_path)

    def postprocess_model_xml(self, xml_str):
        """
        New version of postprocess model xml that only replaces robosuite file paths if necessary (otherwise
        there is an error with the "max" operation)
        """

        path = os.path.split(robosuite.__file__)[0]
        path_split = path.split("/")

        # replace mesh and texture file paths
        tree = ET.fromstring(xml_str)
        root = tree
        asset = root.find("asset")
        meshes = asset.findall("mesh")
        textures = asset.findall("texture")
        all_elements = meshes + textures

        for elem in all_elements:
            old_path = elem.get("file")
            if old_path is None:
                continue

            old_path_split = old_path.split("/")
            # maybe replace all paths to robosuite assets
            check_lst = [
                loc for loc, val in enumerate(old_path_split) if val == "robosuite"
            ]
            if len(check_lst) > 0:
                ind = max(check_lst)  # last occurrence index
                new_path_split = path_split + old_path_split[ind + 1 :]
                new_path = "/".join(new_path_split)
                elem.set("file", new_path)

        return ET.tostring(root, encoding="utf8").decode("utf8")

    def _get_geoms(self, root, _parent=None):
        """
        Helper function to recursively search through element tree starting at @root and returns
        a list of (parent, child) tuples where the child is a geom element

        Args:
            root (ET.Element): Root of xml element tree to start recursively searching through

            _parent (ET.Element): Parent of the root element tree. Should not be used externally; only set
                during the recursive call

        Returns:
            list: array of (parent, child) tuples where the child element is a geom type
        """
        geom_pairs = super(MJCFSiteObject, self)._get_geoms(root=root, _parent=_parent)

        # modify geoms according to the attributes
        for i, (parent, element) in enumerate(geom_pairs):
            element.set("solref", array_to_string(self.solref))
            element.set("solimp", array_to_string(self.solimp))
            element.set("density", str(self.density))
            element.set("friction", array_to_string(self.friction))
            if self.margin is not None:
                element.set("margin", str(self.margin))

            if (self.rgba is not None) and (element.get("group") == "1"):
                element.set("rgba", array_to_string(self.rgba))

            if self.priority is not None:
                # set high priorit
                element.set("priority", str(self.priority))

        return geom_pairs

    @property
    def horizontal_radius(self):
        horizontal_radius_site = self.worldbody.find(
            "./body/site[@name='{}horizontal_radius_site']".format(self.naming_prefix)
        )
        site_values = string_to_array(horizontal_radius_site.get("pos"))
        return np.linalg.norm(site_values[0:2])

    def get_bbox_points(self, trans=None, rot=None):
        """
        Get the full 8 bounding box points of the object
        rot: a rotation matrix
        """
        bbox_offsets = []

        bottom_offset = self.bottom_offset
        top_offset = self.top_offset
        horizontal_radius_site = self.worldbody.find(
            "./body/site[@name='{}horizontal_radius_site']".format(self.naming_prefix)
        )
        horiz_radius = string_to_array(horizontal_radius_site.get("pos"))[:2]

        center = np.mean([bottom_offset, top_offset], axis=0)
        half_size = [horiz_radius[0], horiz_radius[1], top_offset[2] - center[2]]

        bbox_offsets = [
            center + half_size * np.array([-1, -1, -1]),  # p0
            center + half_size * np.array([1, -1, -1]),  # px
            center + half_size * np.array([-1, 1, -1]),  # py
            center + half_size * np.array([-1, -1, 1]),  # pz
            center + half_size * np.array([1, 1, 1]),
            center + half_size * np.array([-1, 1, 1]),
            center + half_size * np.array([1, -1, 1]),
            center + half_size * np.array([1, 1, -1]),
        ]

        if trans is None:
            trans = np.array([0, 0, 0])
        if rot is not None:
            rot = T.quat2mat(rot)
        else:
            rot = np.eye(3)

        points = [(np.matmul(rot, p) + trans) for p in bbox_offsets]
        return points

    @property
    def size(self):
        bottom_offset = self.bottom_offset
        top_offset = self.top_offset
        horizontal_radius_site = self.worldbody.find(
            "./body/site[@name='{}horizontal_radius_site']".format(self.naming_prefix)
        )
        horiz_radius = string_to_array(horizontal_radius_site.get("pos"))[:2]

        size = [
            horiz_radius[0] * 2,
            horiz_radius[1] * 2,
            top_offset[2] - bottom_offset[2],
        ]
        return size


def get_all_models(path):
    """
    Get all model XML files in the given path and its subdirectories.
    """
    return glob.glob(os.path.join(path, '**', 'model.xml'), recursive=True)


def get_bbox_sizes(xml_path):
    obj = MJCFSiteObject("obj", xml_path)
    bottom_offset = obj.bottom_offset
    top_offset = obj.top_offset
    horizontal_radius_site = obj.worldbody.find(
        f"./body/site[@name='{obj.naming_prefix}horizontal_radius_site']"
    )
    horiz_radius = string_to_array(horizontal_radius_site.get("pos"))[:2]

    center = np.mean([bottom_offset, top_offset], axis=0)
    half_size = [horiz_radius[0], horiz_radius[1], top_offset[2] - center[2]]

    return half_size, center


def remove_sites(body):
    """
    Remove <site> elements named bottom_site, top_site, or horizontal_radius_site.
    """
    to_remove = {"bottom_site", "top_site", "horizontal_radius_site"}
    for site in body.findall("site"):
        if site.get("name") in to_remove:
            body.remove(site)


def update_model_xmls(base_path):
    all_models = get_all_models(base_path)
    for model_path in tqdm(all_models):
        
        # Parse with lxml and remove blank text for neat formatting
        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.parse(model_path, parser)
        root = tree.getroot()
        if any(geom.get("name") == "reg_bbox" for geom in root.iter("geom")):
            continue
        print(f"Updating model: {model_path}")

        # Ensure default region block
        default = root.find("default")
        if default is None:
            region_default = etree.fromstring(REGION_GEOM_CLS)
            root.insert(1, region_default)
        else:
            assert default.find("default[@class='region']") is not None, \
                "Default tag should contain region class"

        # Compute bbox and center
        half_size, center = get_bbox_sizes(model_path)
        pos = array_to_string(center)
        size = array_to_string(half_size)

        # Append region geom
        geom = etree.fromstring(GEOM_TEMPLATE.format(pos, size))
        object_body_parent = root.find(".//worldbody/body")
        object_body = root.find(".//worldbody/body/body")
        assert object_body is not None, "Object body not found in the model xml"
        object_body.append(geom)

        # Remove unwanted site elements
        remove_sites(object_body_parent)

        # Indent in-place and write back with pretty_print
        etree.indent(tree, space="  ")
        tree.write(
            model_path,
            pretty_print=True,
            xml_declaration="lightwheel" in model_path,
            encoding='UTF-8'
        )
    

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--base_path", type=str, default=BASE_PATH)
    args = parser.parse_args()
    update_model_xmls(args.base_path)
