import os, tarfile, json, shutil, argparse, h5py
from tqdm import tqdm
import matplotlib.pyplot as plt
import numpy as np

class DataQC:
    def __init__(self, root_path, layout_num=10, per_task_layout_num=10, exclude_path="scene_layout_exclude.json"):
        self.root_path = root_path
        self.stat_path = os.path.join(os.path.dirname(root_path), "task_stat.json")
        self.draw_path = os.path.join(os.path.dirname(root_path), "task_distribution")
        self.remain_path = os.path.join(os.path.dirname(root_path), "task_remain.json")
        self.valid_name_path = os.path.join(os.path.dirname(root_path), "task_valid_name.txt")
        self.valid_data_path = os.path.join(os.path.dirname(root_path), "task_valid_data")
        self.layout_num = layout_num
        self.per_task_layout_num = per_task_layout_num
        self.exclude_path = exclude_path
        self.exclude_task_layout = json.load(open(exclude_path, "r"))
        self.all_num = 0
        self.rng = np.random.RandomState(42)
        self.task_dict = {}
        self.task_layout_num_dict = {}

    def process_data(self):
        for tar_file in tqdm(os.listdir(self.root_path), desc="Processing data"):
            if tar_file.endswith(".tar"):
                self.all_num += 1
                task_name = tar_file.split("-")[0]
                os.makedirs(os.path.join(self.root_path, task_name), exist_ok=True)
                with tarfile.open(os.path.join(self.root_path, tar_file)) as tar:
                    tar.extractall(path=os.path.join(self.root_path, task_name))
            os.remove(os.path.join(self.root_path, tar_file))
    
    def statistics_data(self):
        self.task_dict = {}
        self.task_layout_num_dict = {}
        for task_name in os.listdir(self.root_path):
            if not os.path.isdir(os.path.join(self.root_path, task_name)): continue
            if task_name not in self.task_dict: self.task_dict[task_name] = {}
            for ep_name in tqdm(os.listdir(os.path.join(self.root_path, task_name)), desc=f"Statistics {task_name}"):
                if not os.path.isdir(os.path.join(self.root_path, task_name, ep_name)): continue
                ep_meta_path = os.path.join(self.root_path, task_name, ep_name, "ep_meta.json")
                ep_meta = json.load(open(ep_meta_path, "r"))
                layout_id = ep_meta["layout_id"]
                if layout_id not in self.task_dict[task_name]: self.task_dict[task_name][layout_id] = []
                self.task_dict[task_name][layout_id].append(ep_name)
        for task_name in self.task_dict:
            self.task_layout_num_dict[task_name] = {}
            for layout_id in range(self.layout_num):
                if layout_id in self.task_dict[task_name]:
                    self.task_layout_num_dict[task_name][layout_id] = len(self.task_dict[task_name][layout_id])
                else:
                    self.task_layout_num_dict[task_name][layout_id] = 0
        json.dump(self.task_dict, open(self.stat_path, "w"), indent=4)
        print(f"Statistics data saved to {self.stat_path}")
    
    def filter_invalid_data(self):
        for task_name in tqdm(os.listdir(self.root_path), desc="Filtering invalid data"):
            if not os.path.isdir(os.path.join(self.root_path, task_name)): 
                os.remove(os.path.join(self.root_path, task_name))
                continue
            for ep_name in os.listdir(os.path.join(self.root_path, task_name)):
                if not os.path.isdir(os.path.join(self.root_path, task_name, ep_name)): 
                    os.remove(os.path.join(self.root_path, task_name, ep_name))
                    continue
                f = h5py.File(os.path.join(self.root_path, task_name, ep_name, "ep_demo.hdf5"), "r")
                acs = f["data"][list(f["data"].keys())[0]]["actions"][:]
                if np.any(np.min(acs, axis=0) < -1.0) or np.any(np.max(acs, axis=0) > 1.0):
                    shutil.rmtree(os.path.join(self.root_path, task_name, ep_name))

    def draw_distribution(self):
        os.makedirs(self.draw_path, exist_ok=True)
        if self.task_dict == {}: self.statistics_data()
        for task_name in self.task_dict:
            layout_counts = [0] * self.layout_num  # 初始化[0, layout_num-1]的计数为0
            for layout_id in self.task_dict[task_name]:
                layout_counts[layout_id] = len(self.task_dict[task_name][layout_id])
            plt.figure(figsize=(10, 6))
            bars = plt.bar(range(self.layout_num), layout_counts)
            # 在每个柱形上添加数值标签
            for idx, bar in enumerate(bars):
                height = bar.get_height()
                label = f'{int(height)}'
                if idx in self.exclude_task_layout[task_name]:
                    bar.set_color('yellow')
                    label += '\n(exclude)'
                plt.text(bar.get_x() + bar.get_width()/2., height, label, ha='center', va='bottom')
            plt.xticks(range(self.layout_num))
            plt.title(f"{task_name} Layout Distribution")
            plt.xlabel("Layout ID")
            plt.ylabel("Episode Count")
            plt.savefig(os.path.join(self.draw_path, f"{task_name}.png"))
            plt.close()
    
    def export_valid_data(self):
        all_valid_num = 0
        if self.task_dict == {}: self.statistics_data()
        os.makedirs(os.path.dirname(self.valid_data_path), exist_ok=True)
        os.makedirs(os.path.dirname(self.valid_name_path), exist_ok=True)
        with open(self.valid_name_path, "w") as f:
            for task_name in self.task_layout_num_dict:
                current_valid_num = 0
                os.makedirs(os.path.join(self.valid_data_path, task_name), exist_ok=True)
                # 计算该task中有效的layout数量
                valid_layout_count = sum(1 for lid in range(self.layout_num) if lid not in self.exclude_task_layout[task_name])
                # 计算每个有效layout应该有的数据量
                target_per_layout = (self.per_task_layout_num * self.layout_num) // valid_layout_count
                # 收集该task下所有有效layout的数据
                valid_layouts_data = {}
                for layout_id in self.task_layout_num_dict[task_name]:
                    if layout_id in self.exclude_task_layout[task_name]: continue
                    if layout_id not in self.task_dict[task_name]: valid_layouts_data[layout_id] = []
                    else: valid_layouts_data[layout_id] = self.task_dict[task_name][layout_id]
                # 对每个有效layout进行处理
                left_layout_data = {}
                for layout_id, episodes in valid_layouts_data.items():
                    # 取target_per_layout个数据
                    selected_episodes = episodes[:min(target_per_layout, len(episodes))]
                    left_layout_data[layout_id] = episodes[min(target_per_layout, len(episodes)):]
                    for ep_name in selected_episodes:
                        f.write(f"{task_name}-{ep_name}.tar\n")
                        shutil.copytree(os.path.join(self.root_path, task_name, ep_name), 
                                      os.path.join(self.valid_data_path, task_name, ep_name), 
                                      dirs_exist_ok=True)
                        current_valid_num += 1
                while current_valid_num < self.per_task_layout_num * self.layout_num:
                    left_layout_ids = [k for k in left_layout_data.keys() if len(left_layout_data[k]) > 0]
                    if len(left_layout_ids) == 0: 
                        print(f"No valid layout for {task_name} when balancing data, now have {current_valid_num} data")
                        break
                    rng_layout_id = self.rng.choice(left_layout_ids)
                    if len(left_layout_data[rng_layout_id]) > 0:
                        ep_name = left_layout_data[rng_layout_id].pop(0)
                        f.write(f"{task_name}-{ep_name}.tar\n")
                        shutil.copytree(os.path.join(self.root_path, task_name, ep_name), 
                                      os.path.join(self.valid_data_path, task_name, ep_name), 
                                      dirs_exist_ok=True)
                        current_valid_num += 1

                all_valid_num += current_valid_num
        print(f"Valid data saved to {self.valid_data_path}")
        print(f"Valid name saved to {self.valid_name_path}")
        print(f"Valid data number: {all_valid_num}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--root_path", type=str, default="")
    parser.add_argument("--layout_num", type=int, default=10)
    parser.add_argument("--per_task_layout_num", type=int, default=10)
    parser.add_argument("--exclude_path", type=str, default="scene_layout_exclude.json")
    parser.add_argument("--no_process", action="store_true")
    parser.add_argument("--no_export", action="store_true")
    args = parser.parse_args()
    root_path = args.root_path
    layout_num = args.layout_num
    per_task_layout_num = args.per_task_layout_num
    exclude_path = args.exclude_path
    data_qc = DataQC(root_path, layout_num, per_task_layout_num, exclude_path)
    if not args.no_process: data_qc.process_data()
    data_qc.filter_invalid_data()
    data_qc.statistics_data()
    data_qc.draw_distribution()
    if not args.no_export: data_qc.export_valid_data()