import numpy as np
import os
import sys
import torch

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import gr00t
from gr00t.data.dataset import LeRobotSingleDataset
from gr00t.experiment.data_config import DATA_CONFIG_MAP
from gr00t.model.policy import Gr00tPolicy
from typing import Tuple

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

# change the following paths
MODEL_PATH = os.path.join(CURRENT_DIR, "../", "GR00T-N1-2B")
DATASET_PATH = os.path.join(CURRENT_DIR, "../", "demo_data/robot_sim.PickNPlace")
device = "cuda" if torch.cuda.is_available() else "cpu"


def load_model(
    config: str = "gr1_arms_only",
    tag: str = "gr1",
    model_path: str = MODEL_PATH,
    dataset_path: str = DATASET_PATH,
    device: str = device,
) -> Tuple[LeRobotSingleDataset, Gr00tPolicy]:
    """Load N1 Dataset and Model

    Args:
        config (str): data config for robot. Defaults to "gr1_arms_only".
        tag (str): the used robot. Defaults to "gr1".
        model_path (str): path of the gr00t N1 model. Defaults to MODEL_PATH.
        dataset_path (str): dataset with the LeRobotDataset-V2 format. Defaults to DATASET_PATH.
        device (str): cuda or cpu. Defaults to device.

    Returns:
        dataset class and policy class
            ->dataset: LeRobotSingleDataset class
            ->policy: Gr00tPolicy class
    """

    data_config = DATA_CONFIG_MAP[config]
    modality_config = data_config.modality_config()
    modality_transform = data_config.transform()

    # Load Policy
    policy = Gr00tPolicy(
        model_path=model_path,
        embodiment_tag=tag,
        modality_config=modality_config,
        modality_transform=modality_transform,
        device=device,
    )
    
    modality_config = policy.modality_config

    # Loading dataset
    dataset = LeRobotSingleDataset(
        dataset_path=dataset_path,
        modality_configs=modality_config,
        video_backend="decord",
        video_backend_kwargs=None,
        transforms=None,  # We'll handle transforms separately through the policy
        embodiment_tag=tag,
    )

    return (dataset, policy)